@@ -1,6 +1,6 @@
 /* _NVRM_COPYRIGHT_BEGIN_
  *
- * Copyright 2001-2002 by NVIDIA Corporation.  All rights reserved.  All
+ * Copyright 2001-2011 by NVIDIA Corporation.  All rights reserved.  All
  * information contained herein is proprietary and confidential to NVIDIA
  * Corporation.  Any use, reproduction, or disclosure without the written
  * permission of NVIDIA Corporation is prohibited.
@@ -13,32 +13,28 @@
 #include "nv.h"
 #include "nv-freebsd.h"
 
-
-/*
- * The NVIDIA kernel module's malloc identifier, needed for both tracking
- * and actual allocation/freeing purposes. M_NVIDIA is declared elsewhere
- * to make it known to other parts of the kernel module (nv-freebsd.h).
- */
-
 MALLOC_DEFINE(M_NVIDIA, "nvidia", "NVIDIA memory allocations");
-
-
-#define MAX_ERROR_STRING    256
-
-/*
- * The core resource manager's favorite source of memory, this routine is
- * called from different contexts, including ISRs. This means that it can
- * not be allowed to sleep when memory is low.
- */
+TASKQUEUE_DEFINE_THREAD(nvidia);
 
 RM_STATUS NV_API_CALL os_alloc_mem(
     void **address,
-    NvU32 size
+    NvU64 size
 )
 {
+    unsigned long alloc_size;
+
+    /*
+     * malloc takes an input of unsigned long (8 bytes in x64, 4 bytes in x86).
+     * To avoid truncation and wrong allocation, below check is required.
+     */
+    alloc_size = size;
+
+    if (alloc_size != size)
+        return RM_ERR_INVALID_LIMIT;
+
     /* XXX Fix me? (malloc flags) */
-    *address = malloc(size, M_NVIDIA, M_NOWAIT | M_ZERO);
-    return *address ? RM_OK : RM_ERROR;
+    *address = malloc(alloc_size, M_NVIDIA, M_NOWAIT | M_ZERO);
+    return *address ? RM_OK : RM_ERR_NO_FREE_MEM;
 }
 
 void NV_API_CALL os_free_mem(void *address)
@@ -102,9 +98,10 @@ RM_STATUS NV_API_CALL os_delay_us(NvU32 MicroSeconds)
     return RM_OK;
 }
 
-NvU32 NV_API_CALL os_get_cpu_frequency(void)
+NvU64 NV_API_CALL os_get_cpu_frequency(void)
 {
-    return ((tsc_freq + 4999) / 1000000);
+    /* round up by 4999 before division by 1000000 in osGetCpuFrequency() */
+    return (tsc_freq + 4999);
 }
 
 NvU32 NV_API_CALL os_get_current_process(void)
@@ -206,11 +203,6 @@ void* NV_API_CALL os_map_kernel_space(
         return NULL;
 #endif
 
-    if (start & PAGE_MASK)
-        return NULL;
-
-    size = NV_ALIGN_UP(size, PAGE_SIZE);
-
     switch (mode) {
         case NV_MEMORY_CACHED:
             map_mode = PAT_WRITE_BACK;
@@ -239,10 +231,30 @@ void NV_API_CALL os_unmap_kernel_space(
     pmap_unmapdev((vm_offset_t)address, size);
 }
 
+void* NV_API_CALL os_map_user_space(
+    NvU64   start,
+    NvU64   size_bytes,
+    NvU32   mode,
+    NvU32   protect,
+    void  **priv_data
+)
+{
+    return (void *)(NvUPtr)start;
+}
+
+void NV_API_CALL os_unmap_user_space(
+    void  *address,
+    NvU64  size,
+    void  *priv_data
+)
+{
+}
+
 RM_STATUS NV_API_CALL os_set_mem_range(
     NvU64 start,
     NvU64 size,
-    NvU32 mode
+    NvU32 mode,
+    NvS32 *memory_token
 )
 {
     int arg;
@@ -263,7 +275,8 @@ RM_STATUS NV_API_CALL os_set_mem_range(
 
 RM_STATUS NV_API_CALL os_unset_mem_range(
     NvU64 start,
-    NvU64 size
+    NvU64 size,
+    NvS32 memory_token
 )
 {
     int arg;
@@ -311,6 +324,46 @@ RM_STATUS NV_API_CALL os_schedule(void)
     return RM_ERR_NOT_SUPPORTED;
 }
 
+static void os_execute_work_item(void *context, int pending)
+{
+    nvidia_work_t *work = (nvidia_work_t *)context;
+    nv_stack_t *sp = NULL;
+
+    NV_UMA_ZONE_ALLOC_STACK(sp);
+    if (sp == NULL) {
+        nv_printf(NV_DBG_ERRORS, "NVRM: failed to allocate stack!\n");
+        return;
+    }
+
+    rm_execute_work_item(sp, work->data);
+    os_free_mem((void *)work);
+
+    NV_UMA_ZONE_FREE_STACK(sp);
+}
+
+RM_STATUS NV_API_CALL os_queue_work_item(void *data)
+{
+    RM_STATUS status;
+    nvidia_work_t *work;
+
+    status = os_alloc_mem((void **)&work, sizeof(nvidia_work_t));
+    if (status != RM_OK)
+        return status;
+
+    work->data = data;
+
+    TASK_INIT(&work->task, 0, os_execute_work_item, (void *)work);
+    taskqueue_enqueue(taskqueue_nvidia, &work->task);
+
+    return RM_OK;
+}
+
+RM_STATUS NV_API_CALL os_flush_work_queue(void)
+{
+    taskqueue_run(taskqueue_nvidia);
+    return RM_OK;
+}
+
 void NV_API_CALL os_dbg_set_level(NvU32 new_debuglevel)
 {
     cur_debuglevel = new_debuglevel;
@@ -328,6 +381,7 @@ void NV_API_CALL out_string(const char *message)
     printf("%s", message);
 }
 
+#define MAX_ERROR_STRING 512
 static char nv_error_string[MAX_ERROR_STRING];
 
 int NV_API_CALL nv_printf(
@@ -342,7 +396,7 @@ int NV_API_CALL nv_printf(
 
     if (debuglevel >= ((cur_debuglevel >> 4) & 3)) {
         va_start(arglist, format);
-        chars_written = vsprintf(message, format, arglist);
+        chars_written = vsnprintf(message, sizeof(nv_error_string), format, arglist);
         va_end(arglist);
         printf("%s", message);
     }
@@ -350,9 +404,9 @@ int NV_API_CALL nv_printf(
     return chars_written;
 }
 
-int NV_API_CALL nv_snprintf(
+NvS32 NV_API_CALL os_snprintf(
     char *buf,
-    unsigned int size,
+    NvU32 size,
     const char *fmt,
     ...
 )
@@ -367,10 +421,11 @@ int NV_API_CALL nv_snprintf(
     return chars_written;
 }
 
-void NV_API_CALL nv_os_log(
-    int loglevel,
+void NV_API_CALL os_log_error(
+    NvU32 log_level,
+    NvU32 error_number,
     const char *fmt,
-    void *ap
+    va_list ap
 )
 {
     int l;
@@ -410,10 +465,7 @@ RM_STATUS NV_API_CALL os_memcpy_from_user(
     NvU32 length
 )
 {
-    if (src < (void *) VM_MAXUSER_ADDRESS)
-        return copyin(src, dst, length)  ? RM_ERR_INVALID_POINTER : RM_OK;
-
-    return os_mem_copy(dst, src, length) ? RM_ERR_INVALID_POINTER : RM_OK;
+    return copyin(src, dst, length)  ? RM_ERR_INVALID_POINTER : RM_OK;
 }
 
 RM_STATUS NV_API_CALL os_memcpy_to_user(
@@ -422,10 +474,7 @@ RM_STATUS NV_API_CALL os_memcpy_to_user(
     NvU32 length
 )
 {
-    if (dst < (void *) VM_MAXUSER_ADDRESS)
-        return copyout(src, dst, length) ? RM_ERR_INVALID_POINTER : RM_OK;
-
-    return os_mem_copy(dst, src, length) ? RM_ERR_INVALID_POINTER : RM_OK;
+    return copyout(src, dst, length) ? RM_ERR_INVALID_POINTER : RM_OK;
 }
 
 void* NV_API_CALL os_mem_set(
@@ -499,6 +548,17 @@ RM_STATUS NV_API_CALL os_flush_cpu_cache(void)
 }
 #endif
 
+RM_STATUS NV_API_CALL os_flush_user_cache(
+    NvU64 start,
+    NvU64 end,
+    NvU64 physStart,
+    NvU64 physEnd, 
+    NvU32 flags
+)
+{
+    return RM_ERR_NOT_SUPPORTED;
+}
+
 static void sfence_action_func(void *arg)
 {
     __asm__ __volatile__("sfence": : :"memory");
@@ -520,6 +580,14 @@ RM_STATUS NV_API_CALL os_clear_smp_barrier(void)
     return RM_ERR_NOT_SUPPORTED;
 }
 
+void NV_API_CALL os_disable_preemption(void)
+{
+}
+
+void NV_API_CALL os_enable_preemption(void)
+{
+}
+
 struct os_mutex {
     nv_stack_t *sp;
     struct mtx mutex_mtx;
@@ -554,7 +622,7 @@ RM_STATUS NV_API_CALL os_alloc_sema(void **semaphore)
     return RM_OK;
 }
 
-RM_STATUS NV_API_CALL os_free_sema(void *semaphore)
+void NV_API_CALL os_free_sema(void *semaphore)
 {
     struct os_mutex *mtx = semaphore;
     nv_stack_t *sp;
@@ -566,8 +634,6 @@ RM_STATUS NV_API_CALL os_free_sema(void *semaphore)
     cv_destroy(&mtx->mutex_cv);
 
     os_free_mem(semaphore);
-
-    return RM_OK;
 }
 
 RM_STATUS NV_API_CALL os_acquire_sema(void *semaphore)
@@ -623,11 +689,149 @@ BOOL NV_API_CALL os_is_acquired_sema(void *semaphore)
     return (mtx->refcnt < 1);
 }
 
+
+RM_STATUS NV_API_CALL os_alloc_mutex(void **mutex)
+{
+    RM_STATUS status;
+    struct sx *sx;
+
+    status = os_alloc_mem((void **)&sx, sizeof(struct sx));
+    if (status != RM_OK)
+        return status;
+
+    sx_init(sx, "os.lock_sx");
+    *mutex = (void *)sx;
+
+    return RM_OK;
+}
+
+void NV_API_CALL os_free_mutex(void *mutex)
+{
+    struct sx *sx = mutex;
+
+    if (sx != NULL) {
+        sx_destroy(sx);
+        os_free_mem(sx);
+    }
+}
+
+RM_STATUS NV_API_CALL os_acquire_mutex(void *mutex)
+{
+    struct sx *sx = mutex;
+
+    if (__NV_ITHREAD())
+        return RM_ERR_INVALID_REQUEST;
+
+    sx_xlock(sx);
+
+    return RM_OK;
+}
+
+RM_STATUS NV_API_CALL os_cond_acquire_mutex(void *mutex)
+{
+    struct sx *sx = mutex;
+
+    if (__NV_ITHREAD())
+        return RM_ERR_INVALID_REQUEST;
+
+    if (sx_try_xlock(sx) == 0)
+        return RM_ERR_TIMEOUT_RETRY;
+
+    return RM_OK;
+}
+
+void NV_API_CALL os_release_mutex(void *mutex)
+{
+    struct sx *sx = mutex;
+
+    sx_xunlock(sx);
+}
+
+struct os_semaphore {
+    struct mtx mutex_mtx;
+    struct cv mutex_cv;
+    NvS32  count;
+    NvS32  limit;
+};
+
+void* NV_API_CALL os_alloc_semaphore(NvU32 initialValue, NvU32 limit)
+{
+    RM_STATUS status;
+    struct os_semaphore *os_sema;
+
+    status = os_alloc_mem((void **)&os_sema, sizeof(struct os_semaphore));
+    if (status != RM_OK) {
+        nv_printf(NV_DBG_ERRORS, "NVRM: failed to allocate semaphore!\n");
+        return NULL;
+    }
+
+    mtx_init(&os_sema->mutex_mtx, "os.sema_mtx", NULL, MTX_DEF);
+    cv_init(&os_sema->mutex_cv, "os.sema_cv");
+
+    os_sema->count = initialValue;
+    os_sema->limit = limit;
+
+    return (void *)os_sema;
+}
+
+void NV_API_CALL os_free_semaphore(void *semaphore)
+{
+    struct os_semaphore *os_sema = (struct os_semaphore *)semaphore;
+
+    mtx_destroy(&os_sema->mutex_mtx);
+    cv_destroy(&os_sema->mutex_cv);
+
+    os_free_mem(os_sema);
+}
+
+RM_STATUS NV_API_CALL os_acquire_semaphore(void *semaphore)
+{
+    struct os_semaphore *os_sema = (struct os_semaphore *)semaphore;
+
+    mtx_lock(&os_sema->mutex_mtx);
+    os_sema->count--;
+    if (os_sema->count < 0)
+        cv_wait_unlock(&os_sema->mutex_cv, &os_sema->mutex_mtx);
+    else
+        mtx_unlock(&os_sema->mutex_mtx);
+
+    return RM_OK;
+}
+
+RM_STATUS NV_API_CALL os_release_semaphore(void *semaphore)
+{
+    struct os_semaphore *os_sema = (struct os_semaphore *)semaphore;
+
+    mtx_lock(&os_sema->mutex_mtx);
+    if (os_sema->count < 0) {
+        cv_signal(&os_sema->mutex_cv);
+    }
+    os_sema->count++;
+    mtx_unlock(&os_sema->mutex_mtx);
+
+    return RM_OK;
+}
+
+BOOL NV_API_CALL os_semaphore_may_sleep(void)
+{
+    return (!__NV_ITHREAD());
+}
+
+BOOL NV_API_CALL os_is_isr(void)
+{
+    return (__NV_ITHREAD());
+}
+
 BOOL NV_API_CALL os_pat_supported(void)
 {
     return TRUE;
 }
 
+BOOL NV_API_CALL os_is_efi_enabled(void)
+{
+    return FALSE;
+}
+
 void NV_API_CALL os_register_compatible_ioctl(
     NvU32 cmd,
     NvU32 size
@@ -652,20 +856,47 @@ RM_STATUS NV_API_CALL os_enable_console_access(void)
     return RM_OK;
 }
 
-NvU64 NV_API_CALL os_acquire_spinlock(void *pSema)
+RM_STATUS NV_API_CALL os_alloc_spinlock(void **lock)
 {
-    struct os_mutex *mtx = pSema;
+    RM_STATUS status;
+    struct mtx *mtx;
 
-    mtx_lock(&mtx->mutex_mtx);
+    status = os_alloc_mem((void **)&mtx, sizeof(struct mtx));
+    if (status != RM_OK)
+        return status;
+
+    mtx_init(mtx, "os.lock_mtx", NULL, MTX_DEF);
+
+    *lock = (void *)mtx;
+
+    return RM_OK;
+}
+
+void NV_API_CALL os_free_spinlock(void *lock)
+{
+    struct mtx *mtx = lock;
+
+    if (mtx != NULL) {
+        mtx_assert(mtx, MA_OWNED);
+        mtx_destroy(mtx);
+        os_free_mem(mtx);
+    }
+}
+
+NvU64 NV_API_CALL os_acquire_spinlock(void *lock)
+{
+    struct mtx *mtx = lock;
+
+    mtx_lock(mtx);
 
     return 0;
 }
 
-void NV_API_CALL os_release_spinlock(void *pSema, NvU64 oldIrql)
+void NV_API_CALL os_release_spinlock(void *lock, NvU64 oldIrql)
 {
-    struct os_mutex *mtx = pSema;
+    struct mtx *mtx = lock;
 
-    mtx_unlock(&mtx->mutex_mtx);
+    mtx_unlock(mtx);
 }
 
 RM_STATUS NV_API_CALL os_get_address_space_info(
@@ -677,3 +908,18 @@ RM_STATUS NV_API_CALL os_get_address_space_info(
 {
     return RM_ERR_NOT_SUPPORTED;
 }
+
+RM_STATUS NV_API_CALL os_alloc_smp_barrier(void)
+{
+    return RM_OK;
+}
+
+RM_STATUS NV_API_CALL os_free_smp_barrier(void)
+{
+    return RM_OK;
+}
+
+RM_STATUS NV_API_CALL os_get_version_info(os_version_info * pOsVersionInfo)
+{
+    return RM_ERR_NOT_SUPPORTED;
+}
