@@ -20,9 +20,38 @@
 uma_zone_t nvidia_stack_t_zone;
 static nv_stack_t *__nvidia_init_sp = NULL;
 
+static  void   *nv_find_alloc(struct nvidia_softc *sc, NvU64 address, NvU32 *page_index);
+
 devclass_t nvidia_devclass;
 nv_state_t nvidia_ctl_state;
 
+static  void  *nv_find_alloc(
+    struct nvidia_softc *sc,
+    NvU64  address,
+    NvU32 *page_index
+)
+{
+    NvU32 i = 0;
+    struct nvidia_alloc *at = NULL;
+
+    SLIST_FOREACH(at, &sc->alloc_list, list) {
+        for (i = 0; i < (at->size / PAGE_SIZE); i++) {
+            if (!at->alloc_type_contiguous &&
+                    (address == at->pte_array[i].physical_address))
+                goto found;
+            else if (at->alloc_type_contiguous &&
+                     (address == ((i * PAGE_SIZE) +
+                          at->pte_array[0].physical_address)))
+                goto found;
+        }
+    }
+
+found:
+    if (at != NULL)
+        *page_index = i;
+    return at;
+}
+
 int nvidia_attach(device_t dev)
 {
     int status;
@@ -83,7 +112,7 @@ int nvidia_attach(device_t dev)
             if (!sc->BAR_objects[i])
                 goto failed;
 
-            VM_OBJECT_LOCK(sc->BAR_objects[i]);
+            VM_OBJECT_WLOCK(sc->BAR_objects[i]);
             switch (i) {
                 case NV_GPU_BAR_INDEX_FB:
                     vm_object_set_memattr(sc->BAR_objects[i],
@@ -95,7 +124,7 @@ int nvidia_attach(device_t dev)
                             VM_MEMATTR_UNCACHEABLE);
                     break;
             }
-            VM_OBJECT_UNLOCK(sc->BAR_objects[i]);
+            VM_OBJECT_WUNLOCK(sc->BAR_objects[i]);
         }
     }
 
@@ -541,9 +570,9 @@ int nvidia_open_dev(
                 goto failed;
             }
 
-            VM_OBJECT_LOCK(sc->UD_object);
+            VM_OBJECT_WLOCK(sc->UD_object);
             vm_object_set_memattr(sc->UD_object, VM_MEMATTR_UNCACHEABLE);
-            VM_OBJECT_UNLOCK(sc->UD_object);
+            VM_OBJECT_WUNLOCK(sc->UD_object);
         }
 
         nv->flags |= NV_FLAG_OPEN;
@@ -718,6 +747,7 @@ NvS32 nv_os_agp_init(
     NvU32 sba  = 0;
     NvU32 rate = (8 | 4 | 2 | 1);
     NvU32 size = 0;
+    NvS32 memory_token = 0;
 
     sc->agp_dev = agp_find_device();
     if (!sc->agp_dev) {
@@ -732,7 +762,7 @@ NvS32 nv_os_agp_init(
     mode = ai.ai_mode;
 
     if (os_set_mem_range(ai.ai_aperture_base, ai.ai_aperture_size,
-                NV_MEMORY_WRITECOMBINED) != RM_OK) {
+                NV_MEMORY_WRITECOMBINED, &memory_token) != RM_OK) {
         /*
          * Failure to set a write-combining range for the AGP aperture is
          * not necessarily a fatal error condition; we don't know at this
@@ -762,7 +792,8 @@ NvS32 nv_os_agp_init(
 
     if (agp_enable(sc->agp_dev, mode) != 0) {
         agp_release(sc->agp_dev);
-        os_unset_mem_range(ai.ai_aperture_base, ai.ai_aperture_size);
+        os_unset_mem_range(ai.ai_aperture_base, ai.ai_aperture_size,
+                           memory_token);
         return EIO;
     }
 
@@ -770,7 +801,8 @@ NvS32 nv_os_agp_init(
 
     if (os_alloc_mem((void **)&bitmap, size) != RM_OK) {
         agp_release(sc->agp_dev);
-        os_unset_mem_range(ai.ai_aperture_base, ai.ai_aperture_size);
+        os_unset_mem_range(ai.ai_aperture_base, ai.ai_aperture_size,
+                           memory_token);
         return EIO;
     }
 
@@ -779,7 +811,8 @@ NvS32 nv_os_agp_init(
     if (rm_set_agp_bitmap(sp, nv, bitmap) != RM_OK) {
         agp_release(sc->agp_dev);
         os_free_mem(bitmap);
-        os_unset_mem_range(ai.ai_aperture_base, ai.ai_aperture_size);
+        os_unset_mem_range(ai.ai_aperture_base, ai.ai_aperture_size,
+                           memory_token);
         return EIO;
     }
 
@@ -803,7 +836,8 @@ NvS32 nv_os_agp_teardown(
     rm_clear_agp_bitmap(sp, nv, &bitmap);
     os_free_mem(bitmap);
 
-    os_unset_mem_range(nv->agp.address, nv->agp.size);
+    /* Third parameter is a no-op on FreeBSD */
+    os_unset_mem_range(nv->agp.address, nv->agp.size, 0);
 
     return 0;
 }
@@ -901,9 +935,9 @@ RM_STATUS NV_API_CALL nv_agp_init(
         return RM_ERR_NO_FREE_MEM;
     }
 
-    VM_OBJECT_LOCK(sc->AGP_object);
+    VM_OBJECT_WLOCK(sc->AGP_object);
     vm_object_set_memattr(sc->AGP_object, VM_MEMATTR_WRITE_COMBINING);
-    VM_OBJECT_UNLOCK(sc->AGP_object);
+    VM_OBJECT_WUNLOCK(sc->AGP_object);
 
     return RM_OK;
 
@@ -1056,7 +1090,7 @@ found:
 
     if (at->alloc_type_contiguous) {
         *private = NULL;
-        return (void *)(at->pte_array[0].virtual_address +
+        return (void *)(NvUPtr)(at->pte_array[0].virtual_address +
                     (i * PAGE_SIZE) + offset);
     }
 
@@ -1066,7 +1100,7 @@ found:
     virtual_address = vm_map_min(kernel_map);
 
     status = vm_map_find(kernel_map, at->object, (i * PAGE_SIZE),
-            &virtual_address, size, VMFS_ANY_SPACE,
+            &virtual_address, size, 0, VMFS_ANY_SPACE,
             (VM_PROT_READ | VM_PROT_WRITE),
             (VM_PROT_READ | VM_PROT_WRITE), 0);
     if (status != KERN_SUCCESS) {
@@ -1181,7 +1215,7 @@ NvS32 nv_alloc_contig_pages(
         return ENOMEM;
     }
 
-    address = kmem_alloc_contig(kernel_map, size, flags, 0,
+    address = kmem_alloc_contig(kmem_arena, size, flags, 0,
             sc->dma_mask, PAGE_SIZE, 0, attr);
     if (!address) {
         status = ENOMEM;
@@ -1211,9 +1245,9 @@ NvS32 nv_alloc_contig_pages(
         goto failed;
     }
 
-    VM_OBJECT_LOCK(at->object);
+    VM_OBJECT_WLOCK(at->object);
     vm_object_set_memattr(at->object, attr);
-    VM_OBJECT_UNLOCK(at->object);
+    VM_OBJECT_WUNLOCK(at->object);
 
     *private = at;
     SLIST_INSERT_HEAD(&sc->alloc_list, at, list);
@@ -1230,8 +1264,7 @@ failed:
         os_flush_cpu_cache();
 
     if (at->pte_array[0].virtual_address != NULL) {
-        kmem_free(kernel_map,
-                at->pte_array[0].virtual_address, at->size);
+        kmem_free(kmem_arena, at->pte_array[0].virtual_address, at->size);
         malloc_type_freed(M_NVIDIA, at->size);
     }
 
@@ -1257,8 +1290,7 @@ NvS32 nv_free_contig_pages(
     if (at->attr != VM_MEMATTR_WRITE_BACK)
         os_flush_cpu_cache();
 
-    kmem_free(kernel_map, at->pte_array[0].virtual_address,
-            at->size);
+    kmem_free(kmem_arena, at->pte_array[0].virtual_address, at->size);
     malloc_type_freed(M_NVIDIA, at->size);
 
     free(at->pte_array, M_NVIDIA);
@@ -1321,7 +1353,7 @@ NvS32 nv_alloc_system_pages(
     }
 
     for (i = 0; i < count; i++) {
-        address = kmem_alloc_contig(kernel_map, PAGE_SIZE, flags, 0,
+        address = kmem_alloc_contig(kmem_arena, PAGE_SIZE, flags, 0,
                 sc->dma_mask, PAGE_SIZE, 0, attr);
         if (!address) {
             status = ENOMEM;
@@ -1354,9 +1386,9 @@ NvS32 nv_alloc_system_pages(
         goto failed;
     }
 
-    VM_OBJECT_LOCK(at->object);
+    VM_OBJECT_WLOCK(at->object);
     vm_object_set_memattr(at->object, attr);
-    VM_OBJECT_UNLOCK(at->object);
+    VM_OBJECT_WUNLOCK(at->object);
 
     *private = at;
     SLIST_INSERT_HEAD(&sc->alloc_list, at, list);
@@ -1375,8 +1407,7 @@ failed:
     for (i = 0; i < count; i++) {
         if (at->pte_array[i].virtual_address == 0)
             break;
-        kmem_free(kernel_map,
-                at->pte_array[i].virtual_address, PAGE_SIZE);
+        kmem_free(kmem_arena, at->pte_array[i].virtual_address, PAGE_SIZE);
         malloc_type_freed(M_NVIDIA, PAGE_SIZE);
     }
 
@@ -1405,8 +1436,7 @@ NvS32 nv_free_system_pages(
         os_flush_cpu_cache();
 
     for (i = 0; i < count; i++) {
-        kmem_free(kernel_map,
-                at->pte_array[i].virtual_address, PAGE_SIZE);
+        kmem_free(kmem_arena, at->pte_array[i].virtual_address, PAGE_SIZE);
         malloc_type_freed(M_NVIDIA, PAGE_SIZE);
     }
 
@@ -1685,14 +1715,15 @@ int nvidia_mmap_dev_single(
     nv_stack_t *sp;
     RM_STATUS rmStatus;
     struct nvidia_softc *sc = nv->os_state;
-    NvU32 i, prot;
+    NvU32 i = 0;
+    NvU32 prot;
 
     sp = sc->api_sp;
 
     NV_PCI_CHECK_CONFIG_SPACE(sp, nv, TRUE, TRUE, TRUE);
 
     rmStatus = rm_validate_mmap_request(sp, nv, *offset, size, &prot);
-    if (rmStatus != RM_OK)
+    if ((rmStatus != RM_OK) && (NULL == nv_find_alloc(sc, *offset, &i)))
         return EINVAL;
 
     if (IS_UD_OFFSET(nv, *offset, size)) {
@@ -1717,22 +1748,8 @@ int nvidia_mmap_dev_single(
         return 0;
     }
 
-    SLIST_FOREACH(at, &sc->alloc_list, list) {
-        for (i = 0; i < (at->size / PAGE_SIZE); i++) {
-            if (!at->alloc_type_contiguous &&
-                    (*offset == at->pte_array[i].physical_address))
-                goto found;
-            else if (at->alloc_type_contiguous &&
-                     (*offset == ((i * PAGE_SIZE) +
-                          at->pte_array[0].physical_address)))
-                goto found;
-        }
-    }
-
-    return EINVAL;
-
-found:
-    if ((size + (i * PAGE_SIZE)) > at->size)
+    at = nv_find_alloc(sc, *offset, &i);
+    if ((NULL == at) || (size + (i * PAGE_SIZE)) > at->size)
         return EINVAL;
 
     vm_object_reference(at->object);
